from rfc822 import parseaddr

from zope.schema.vocabulary import SimpleVocabulary
from zope.schema.fieldproperty import FieldProperty
from zope.interface import implements
from zope.app.component.hooks import setSite
from zope.app import zapi
from zope.app.component.interfaces import IPossibleSite
from zope.app.container.interfaces import IObjectRemovedEvent
from zope.app.container.interfaces import IObjectAddedEvent
from zope.app.container.interfaces import IContainerModifiedEvent
from OFS.interfaces import IObjectWillBeAddedEvent
from OFS.event import ObjectWillBeRemovedEvent
from zope.app.container.contained import ObjectAddedEvent
from zope.event import notify

from Products.Five.site.localsite import FiveSite
from Products.Five.site.localsite import enableLocalSiteHook

from Products.CMFCore import permissions as CMFPermissions
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.DynamicType import DynamicType
from Products.CMFCore.CMFCatalogAware import CMFCatalogAware

from Products.listen import permissions, config
from Products.listen.lib.browser_utils import encode

from mailboxer_list import MailBoxerMailingList
from Products.listen.interfaces import IMailingList
from Products.listen.interfaces import ISubscriptionList
from Products.listen.interfaces import ISearchableArchive

from plone.mail import decode_header

def addMailingList(self, id, title=u''):
    """ Add a Document """
    o = MailingList(id, title)
    self._setObject(id,o)


class MailingList(DynamicType, CMFCatalogAware, MailBoxerMailingList, FiveSite):
    """ A Mailng list implementation built on top of a customized MailBoxer.

    Let's see how this works:

    Let's set up the test rig, so that we can get our archived vocabulary
    working (the interface won't validate without the vocabulary):

        >>> import Products.Five
        >>> from Products.Five import zcml
        >>> zcml.load_config('meta.zcml', Products.Five)
        >>> from Products.listen.content import tests
        >>> zcml.load_config('configure.zcml', tests)

    First we create a new instance to play with:

        >>> from Products.listen.content import MailingList
        >>> test_list = MailingList(id="test_list", title=u"My Title")

    Make sure the object and class implement our interface:

        >>> from Products.listen.interfaces import IMailingList
        >>> from zope.interface.verify import verifyObject
        >>> from zope.interface.verify import verifyClass
        >>> verifyObject(IMailingList, test_list)
        True
        >>> verifyClass(IMailingList, MailingList)
        True

    Set some basic default properties:

        >>> test_list.description = u'My Description'
        >>> test_list.mailto = 'address@example.com'

    Verify that we have our Title and Description for the UI:

        >>> test_list.Title()
        'My Title'
        >>> test_list.setTitle(u'My New Title')
        >>> test_list.title
        u'My New Title'
        >>> test_list.Description()
        'My Description'
        >>> test_list.setDescription(u'My New Description')
        >>> test_list.description
        u'My New Description'


    Check property validation:
    --------------------------

    The fields should reject inappropriate input:

        >>> from zope.schema._bootstrapinterfaces import RequiredMissing, WrongType, InvalidValue
        >>> try:
        ...     test_list.description = ['My Description']
        ... except WrongType, e:
        ...     print e
        ...
        (['My Description'], <type 'unicode'>)

    We need unicode for description and ASCII for the list address:
        >>> try:
        ...     test_list.description = 'My Description'
        ... except WrongType, e:
        ...     print e
        ...
        ('My Description', <type 'unicode'>)
        >>> try:
        ...     test_list.mailto = u'address@example.com'
        ... except WrongType, e:
        ...     print e
        ...
        (u'address@example.com', <type 'str'>)
        >>> try:
        ...     test_list.mailto = 'ad\xF4dress@example.com'
        ... except InvalidValue:
        ...     print "Invalid"
        ...
        Invalid


    Let's check our backwards compatibility properties
    --------------------------------------------------

    Values entered in moderated and owner should be combined into
    MailBoxer's moderator property:

        >>> test_list.moderators = (u'addr1@example.com',u'addr2@example.com')
        >>> test_list.list_owner = u'owner@example.com'
        >>> test_list.moderator
        (u'owner@example.com', u'addr1@example.com', u'addr2@example.com')
        >>> test_list.moderator = (u'addr1@example.com',u'addr2@example.com')
        >>> test_list.list_owner
        u'addr1@example.com'
        >>> test_list.moderators
        (u'addr2@example.com',)
        >>> try:
        ...     test_list.moderator = ()
        ... except RequiredMissing:
        ...     print "Required Missing"
        ...
        Required Missing
        >>> test_list.moderator = (u'addr1@example.com',)
        >>> test_list.moderators
        ()
        >>> test_list.list_owner
        u'addr1@example.com'

    The MailBoxer required unclosed property should mirror the more
    sensibly named closed property, and they should both always return
    booleans:

        >>> test_list.closed = True
        >>> test_list.unclosed
        False
        >>> test_list.closed = False
        >>> test_list.unclosed
        True
        >>> test_list.unclosed = 0
        >>> test_list.unclosed
        False
        >>> test_list.closed
        True
        >>> test_list.unclosed = 'I am True'
        >>> test_list.unclosed
        True
        >>> test_list.closed
        False

    The MailBoxer maillist property is now implemented using an adapter
    providing the ISubscriptionList interface.  We create a dumb partial
    implementation of ISubscriptionList in /tests/ to test (would be nice
    if we could use ztapi.provideAdapter here):

    Without an adapter provided, this should fail
        >>> try:
        ...     test_list.maillist
        ... except TypeError, e:
        ...     print e
        ...
        ('Could not adapt', <MailingList at test_list>, <InterfaceClass Products.listen.interfaces.mailinglist.ISubscriptionList>)

    Attach the adapter through configuration:
        >>> configure_zcml = '''
        ... <configure xmlns="http://namespaces.zope.org/zope">
        ...    <adapter
        ...        for="Products.listen.interfaces.IMailingList"
        ...        provides="Products.listen.interfaces.ISubscriptionList"
        ...        factory="Products.listen.content.tests.Subscriptions"
        ...        />
        ... </configure>'''
        >>> zcml.load_string(configure_zcml)

    It should work now:
        >>> test_list.maillist
        (u'addr1@example.com', u'addr2@example.com')

    Some interface methods are tested in the MailBoxerMailingList
    parent class.
    """

    implements(IMailingList, IPossibleSite)

    portal_type = "Mailing List"
    meta_type = "MailingList"
    default_view = 'mailinglist_view'
    content_icon   = 'mailboxer_icon.png'

    # Provide our basic properties with schema based validation
    title = FieldProperty(IMailingList['title'])
    description = FieldProperty(IMailingList['description'])
    mailto = FieldProperty(IMailingList['mailto'])
    moderated = FieldProperty(IMailingList['moderated'])
    archived = FieldProperty(IMailingList['archived'])
    closed = FieldProperty(IMailingList['closed'])
    list_owner = FieldProperty(IMailingList['list_owner'])
    moderators = FieldProperty(IMailingList['moderators'])

    # Define some DC fields for plone ui purposes
    def Title(self):
        """The portal catalog's text indices need encoded strings, or else
           they blow up.  It's got to be all unicode or no unicode (or
           perhaps TXNG)."""
        return encode(self.title, self)

    def setTitle(self, val):
        encoding = 'utf-8'
        # Some plone-iness here to get site encoding
        putils = getToolByName(self, 'plone_tool', None)
        if putils is not None:
            encoding = putils.getSiteEncoding()
        self.title = val.decode(encoding)

    def Description(self):
        """The portal catalog's text indices need encoded strings, or else
           they blow up.  It's got to be all unicode or no unicode (or
           perhaps TXNG)."""
        return encode(self.description or u'', self)

    def setDescription(self, val):
        self.description = val or u''

    def SearchableText(self):
        return self.Title() + ' ' + self.Description()

    # Define a moderator property for backwards compatibility with MailBoxer
    # which combines owner and moderators
    def _setModerator(self, val):
        if tuple(val) != self.moderator:
            if val:
                self.list_owner = val[0]
                self.moderators = val[1:]
            else:
                #An owner is required, so this should raise an error
                self.list_owner = None
                self.moderators = ()

    def _getModerator(self):
        mods = self.moderators
        owner = self.list_owner
        if owner:
            return (owner,) + mods
        else:
            return mods

    moderator = property(fget =_getModerator, fset = _setModerator)

    # Implement the awkward unclose property for backwards compatibility
    def _setUnClosed(self, val):
        self.closed = not(val)

    def _getUnClosed(self):
        return not self.closed

    unclosed = property(fget=_getUnClosed, fset=_setUnClosed)

    # MailBoxer wants to use a simple property to store/retrieve the
    # subscribers, we want to use adapters.  We will provide the property
    # only for backwards compatibility with MailBoxer internals.
    def _getMailList(self):
        subscription = ISubscriptionList(self)
        return subscription.subscribers

    def _setMailList(self, value):
        subscription = ISubscriptionList(self)
        subscription.subscribers = value

    maillist = property(fget=_getMailList, fset=_setMailList)


    def _getAllowedSenders(self):
        return tuple(self.subscriber_data)
    allowed_senders = property(fget=_getAllowedSenders)


    def manage_afterAdd(self, item, container):
        # This should be done with events for zope 2.9
        CMFCatalogAware.manage_afterAdd(self, item, container)
        # This only needs to be done when added via a z3 form, because object
        # creation will not be complete in the CMF way of doing things.
        if not self.getPortalTypeName():
            ttool = getToolByName(self, 'portal_types')
            fti = getattr(ttool, self.portal_type)
            fti._finishConstruction(self)
        # Make this possible site a real site
        try:
            enableLocalSiteHook(item)
        except TypeError:
            pass
        setSite(item)
        # Add search utility (catalog) if it's not already present
        sm = self.getSiteManager()
        search = zapi.createObject('listen.SearchUtilityFactory', self.catalog)
        try:
            sm.registerUtility(ISearchableArchive, search)
        except ValueError:
            # Fail if the utility already exists
            pass
        MailBoxerMailingList.manage_afterAdd(self, item, container)
        if not config.HAS_CONTAINER_EVENTS:
            notify(ObjectAddedEvent(self, container, self.getId()))

    def manage_beforeDelete(self, item, container):
        if not config.HAS_CONTAINER_EVENTS:
            notify(ObjectWillBeRemovedEvent(self, container, self.getId()))

    # Use allowedContentTypes to restrict the addable types:
    def allowedContentTypes( self ):
        return []

    # Don't show the contents tab here or on any children:
    def displayContentsTab(self):
        return False

    def fixupMessages(self):
        """A method to rearchive everything, and fix up encoding issues"""
        # don't bother with this stuff on removal, it's pointless
        search = zapi.getUtility(ISearchableArchive, context=self)
        # XXX: Assumes a ZCatalog based utility, we need to rebuild the whole
        # catalog after a move
        # Clear the catalog
        search.manage_catalogClear()
        # Our fix up and index method
        def index_meth(obj, *args):
            # body, address, title, and subject are unicode
            if not isinstance(obj.from_addr, unicode):
                obj.from_addr = decode_header(obj.from_addr)
            if not isinstance(obj.subject, unicode):
                obj.subject = decode_header(obj.subject)

            sender_name = parseaddr(obj.from_addr)[0]

            title_subj = len(obj.subject) > 20 and (obj.subject[:20]+' ...') \
                             or obj.subject
            if sender_name:
                obj.title = u"%s / %s" % (title_subj, sender_name)
            else:
                obj.title= title_subj
            if not isinstance(obj.body, unicode):
                # Default western european encoding, because other info has
                # been lost at this point.
                obj.body = obj.body.decode('iso-8859-1', 'replace')
            search.catalog_object(obj, *args)
        path = '/'.join(self.getPhysicalPath())
        search.ZopeFindAndApply(self, obj_metatypes=('MailingListMessage',),
                                apply_func=index_meth,
                                apply_path=path,
                                search_sub=1)
        # We refresh to catch the inter-message dependencies due to threading
        search.refreshCatalog()
        return "Done."

# Event listener to catalog mail_obj using the ISearchableArchive utility
def catalogMailBoxerMail(mail_obj, event):
    # We need to pass a context because the lookup tool leaves us in the
    # Application context.
    util = zapi.getUtility(ISearchableArchive, context=mail_obj)
    util.indexNewMessage(mail_obj)

# Event listeners for doing the normal catalog unregistration, etc.
def MailingListWillBeMoved(ml, event):
    if not IObjectWillBeAddedEvent.providedBy(event):
        ml.unindexObject()

def MailingListMoved(ml, event):
    if not IObjectRemovedEvent.providedBy(event) and \
       not IObjectAddedEvent.providedBy(event):
        # don't bother with this on removal, it's pointless, and it's already
        # done on add.
        ml.reindexObject()

def MailingListModified(ml, event):
    if not IContainerModifiedEvent.providedBy(event):
        ml.reindexObject()

# And a vocabulary
def archiveOptionsVocabulary(context):
    archive_options = getattr(context, 'archive_options', None)
    if archive_options is None:
        archive_options = ('not archived', 'plain text', 'with attachments')
    return SimpleVocabulary.fromValues(archive_options)
