import re
from zope.interface import implements
from zope.app.annotation.interfaces import IAnnotations

from persistent.dict import PersistentDict
from AccessControl import Unauthorized
from BTrees.OOBTree import OOBTree, OOTreeSet, OOSet, difference, union

from Products.CMFCore.utils import getToolByName

from Products.listen.interfaces import ISubscriptionList
from Products.listen.interfaces import IMemberSubscriptionList
from Products.listen.interfaces import ICombinedSubscriptionList
from Products.listen.config import PROJECTNAME

email_regex = re.compile('\s*<?(\S+\@\S+)>?\s*')

class BaseSubscriptionList(object):
    """An adapter implementing ISubscriptionList, requires an IAnnotable
       IMailingList object to adapt.

    Let's give it a try, we'll create an Annotable object and wrap it:
        >>> from Products.listen.content import tests
        >>> from Products.listen.content import BaseSubscriptionList
        >>> obj = tests.DummyAnnotableList()
        >>> slist = BaseSubscriptionList(obj)

    Some basic tests
    ----------------

    We have our Subscription list, let's see how it works:
        >>> slist.subscribers
        ()
        >>> slist.subscribers = (u'email1@example.com',)
        >>> slist.subscribers
        (u'email1@example.com',)

    Let's make sure our changes really stuck to the object
        >>> list2 = BaseSubscriptionList(obj)
        >>> list2.subscribers
        (u'email1@example.com',)

    Test our methods:
        >>> slist.subscribeAddress(u'email2@example.com')
        >>> slist.subscribers
        (u'email1@example.com', u'email2@example.com')
        >>> slist.isSubscribed('email1@example.com')
        True
        >>> slist.isSubscribed('email2@example.com')
        True
        >>> slist.isSubscribed('email3@example.com')
        False
        >>> slist.unsubscribeAddress(u'email1@example.com')
        >>> slist.subscribers
        (u'email2@example.com',)
        >>> slist.unsubscribeAddress(u'email2@example.com')
        >>> slist.subscribers
        ()
        >>> slist.subscribeAddress(u'email1@example.com')
        >>> slist.subscribers
        (u'email1@example.com',)

    Resubscribing should not create duplicate entries:
        >>> slist.subscribeAddress(u'email2@example.com')
        >>> slist.subscribers
        (u'email1@example.com', u'email2@example.com')
        >>> slist.subscribeAddress(u'email2@example.com')
        >>> slist.subscribers
        (u'email1@example.com', u'email2@example.com')

    And unsubscribing an unknown address should do nothing:
        >>> slist.unsubscribeAddress(u'email3@example.com')
        >>> slist.subscribers
        (u'email1@example.com', u'email2@example.com')

    Bounce deletion should delete the passed in addresses, and remove them
    call the IMailingLists bounce removal mechanism (a dummy in this case):
        >>> obj.reset
        False
        >>> slist.deleteBounces((u'email2@example.com',u'email3@example.com'))
        >>> slist.subscribers
        (u'email1@example.com',)
        >>> obj.reset
        True

    It should silently ignore invalid addresses:
        >>> slist.subscribeAddress('dummy2')
        >>> slist.subscribers
        (u'email1@example.com',)
    """

    implements(ISubscriptionList)

    def __init__(self, context):
        super(BaseSubscriptionList, self).__init__(context)
        self.context = context
        self.annotations = IAnnotations(context)
        self.list_data = self.annotations.get(PROJECTNAME, PersistentDict())
        if isinstance(self.list_data, PersistentDict):
            self.list_data = self.annotations[PROJECTNAME] = OOBTree(
                                                               self.list_data)
            self.list_data['subscribers'] = OOTreeSet()
        if isinstance(self.list_data['subscribers'], tuple):
            self.list_data['subscribers'] = OOTreeSet(
                                                self.list_data['subscribers'])
        self.subscriber_data = self.list_data['subscribers']

        norecvmail = self.list_data.get('norecvmail', [])
        if isinstance(norecvmail, list):
            self.list_data['norecvmail'] = OOSet(norecvmail)
        self.norecvmail = self.list_data['norecvmail']

        # XXX: Any subcalsses must use super to delegate to this class
        # see http://fuhm.org/super-harmful/
        super(BaseSubscriptionList, self).__init__(context)

    def _getSubscribers(self):
        return tuple(filter(self.isRecvMail, self.subscriber_data))

    def _setSubscribers(self, values):
        if tuple(self.subscriber_data) != values:
            self.subscriber_data.clear()
            self.subscriber_data.update(values)
        for e in self.norecvmail.keys():
            if e not in self.subscriber_data:
                self.norecvmail.remove(e)

    subscribers = property(fget=_getSubscribers, fset=_setSubscribers)

    def _getAllowedSenders(self):
        return tuple(self.subscriber_data)
    allowed_senders = property(fget=_getAllowedSenders)

    def subscribeAddress(self, email):
        if email_regex.match(email):
            self.subscriber_data.insert(email)
            try:
                self.norecvmail.remove(email)
            except KeyError:
                pass

    def unsubscribeAddress(self, email):
        try:
            self.subscriber_data.remove(email)
        except KeyError:
            pass
        try:
            self.norecvmail.remove(email)
        except KeyError:
            pass

    def isSubscribed(self, email):
        return bool(self.subscriber_data.has_key(email))

    def isRecvMail(self, email):
        if email is None: return False
        return email not in self.norecvmail

    def deleteBounces(self, addresses):
        for email in addresses:
            self.unsubscribeAddress(email)
        # Use the MailingList API here
        self.context.resetBounces(addresses)

    def getAllowedSenders(self):
        return tuple(self.subscriber_data)

    def stopRecvmail(self, email):
        self.norecvmail.insert(email)


class MemberSubscriptionList(BaseSubscriptionList):
    """An adapter implementing IMemberSubscriptionList, requires an IAnnotable
       object to adapt.

    Let's give it a try.  Again we'll create an Annotable object and wrap it.
    Also, we need to create a bogus MembershipTool, with some fake user info:
        >>> from Products.listen.content import tests
        >>> obj = tests.DummyAnnotableList()

    Make a dummy user and set it as the current "authenticated" user in our
    membership tool:
        >>> def_member = tests.DummyMember('dummy1', u'Dummy 1', u'dummy1@example.com')
        >>> mtool = tests.DummyMembershipTool(def_member)
        >>> mtool.dummy1 = def_member

    We make our membership tool acquirable from the list object:
        >>> obj.portal_membership = mtool

    Add some more dummy users just for fun:
        >>> mtool.dummy2 = tests.DummyMember('dummy2', u'Dummy 2', u'dummy2@example.com')
        >>> mtool.dummy3 = tests.DummyMember('dummy3', u'Dummy 3', u'dummy3@example.com')
        >>> mtool.anon = tests.DummyMember('anon', u'', u'')

    Make our subscription list:
        >>> mlist = MemberSubscriptionList(obj)

    Some basic tests
    ----------------
        >>> mlist.subscribers
        ()
        >>> mlist.subscribed_members
        ()

    subscribeMember should subscribe the current "authenticated" member to the
    list, updating both the list of members and email addresses subscribed.
    unSubscribe member removes an entry from both lists:
        >>> mlist.subscribeMember()
        >>> mlist.subscribers
        (u'dummy1@example.com',)
        >>> mlist.subscribed_members
        ('dummy1',)
        >>> mlist.isMemberSubscribed()
        True
        >>> mlist.unsubscribeMember()
        >>> mlist.subscribers
        ()
        >>> mlist.subscribed_members
        ()
        >>> mlist.isMemberSubscribed()
        False

    Subscribing multiple times should not create duplicates, and unsubscribing
    multiple times should cause no problems:
        >>> mlist.subscribeMember()
        >>> mlist.subscribeMember()
        >>> mlist.subscribers
        (u'dummy1@example.com',)
        >>> mlist.unsubscribeMember()
        >>> mlist.unsubscribeMember()
        >>> mlist.subscribers
        ()

    Altering the subscribed_members property directly should update the list
    of subscribed email addresses:
        >>> mlist.subscribed_members = ('dummy1','dummy2')
        >>> mlist.subscribed_members
        ('dummy1', 'dummy2')
        >>> mlist.subscribers
        (u'dummy1@example.com', u'dummy2@example.com')
        >>> mlist.subscribed_members = ('dummy2',)
        >>> mlist.subscribed_members
        ('dummy2',)
        >>> mlist.subscribers
        (u'dummy2@example.com',)

    The current member is not subscribed at the moment, let's make sure:
        >>> mlist.isMemberSubscribed()
        False

    Explicitly adding email addresses to subscribers has no effect on the
    list of members, and emails not corresponding to members are preserved
    when the member list is updated:
        >>> mlist.subscribeAddress(u'a_non_member@example.com')
        >>> mlist.subscribers
        (u'a_non_member@example.com', u'dummy2@example.com')
        >>> mlist.subscribed_members
        ('dummy2',)
        >>> mlist.subscribed_members  = ('dummy1','dummy2')
        >>> mlist.subscribed_members
        ('dummy1', 'dummy2')
        >>> mlist.subscribers
        (u'a_non_member@example.com', u'dummy1@example.com', u'dummy2@example.com')
        >>> mlist.subscribed_members  = ()
        >>> mlist.subscribers
        (u'a_non_member@example.com',)
        >>> mlist.unsubscribeAddress(u'a_non_member@example.com')
        >>> mlist.subscribers
        ()

    Changing the address of a user does not immediately update the email
    address, but updating the list in any way will update all addresses:
        >>> mlist.subscribed_members = ('dummy1', 'dummy2')
        >>> mtool.dummy2 = tests.DummyMember('dummy2', u'Dummy 2 New', u'dummy2_new@example.com')
        >>> mlist.subscribers
        (u'dummy1@example.com', u'dummy2@example.com')
        >>> mlist.subscribeMember()
        >>> mlist.subscribers
        (u'dummy1@example.com', u'dummy2_new@example.com')

    If the current user is Anonymous, the list should be unaltered, same goes
    for unsubscribing:
        >>> mlist.subscribed_members = ('dummy2',)
        >>> mtool.user = mtool.anon
        >>> mlist.subscribeMember()
        >>> mlist.subscribed_members
        ('dummy2',)
        >>> mlist.subscribers
        (u'dummy2_new@example.com',)
        >>> mlist.unsubscribeMember()
        >>> mlist.subscribed_members
        ('dummy2',)
        >>> mlist.subscribers
        (u'dummy2_new@example.com',)

    Removing an address from the list that corresponds to a member should
    remove the member from the member list:
        >>> mlist.subscribed_members
        ('dummy2',)
        >>> mlist.unsubscribeAddress('dummy2_new@example.com')
        >>> mlist.subscribers
        ()
        >>> mlist.subscribed_members
        ()

    This should be true even if the address was removed using another adapter:
        >>> basic_list = BaseSubscriptionList(obj)
        >>> mlist.subscribed_members = ('dummy2',)
        >>> mlist.subscribed_members
        ('dummy2',)
        >>> basic_list.subscribers
        (u'dummy2_new@example.com',)
        >>> basic_list.unsubscribeAddress('dummy2_new@example.com')
        >>> mlist.subscribers
        ()
        >>> mlist.subscribed_members
        ()

    It should silently drop members with invalid addresses (this is a poor
    policy, but will do for now):
        >>> mtool.dummy2 = tests.DummyMember('dummy2', u'Dummy 2 New', u'dummy2_new_at_example.com')
        >>> mlist.subscribed_members = ('dummy2',)
        >>> mlist.subscribed_members
        ()
        >>> mlist.subscribers
        ()

    The Anonymous user is never subscribed:
        >>> mlist.isMemberSubscribed()
        False

    Setting a member to not receive mail should add him to the ignore list
        >>> uid = def_member.getId()
        >>> mlist.subscribeMember(uid)
        >>> mlist.subscribers
        (u'dummy1@example.com',)
        >>> mlist.isMemberRecvMail(uid)
        True
        >>> mlist.stopRecvmailMember(uid)
        >>> mlist.isMemberRecvMail(uid)
        False

    Now setting the member to receive mail again should remove him from the
    ignore list
        >>> mlist.startRecvmailMember(uid)
        >>> mlist.isMemberRecvMail(uid)
        True

    Removing a member from the subscribed list should also remove him from
    the ignore list
        >>> mlist.stopRecvmailMember(uid)
        >>> mlist.isMemberRecvMail(uid)
        False
        >>> mlist.unsubscribeMember(uid)
        >>> mlist.isMemberRecvMail(uid)
        True

    """

    implements(IMemberSubscriptionList)

    # Stores a structure mapping emails to member names and ids
    # this allows looking up the subscriber emails to be very fast(as it is
    # done for every mail to the list).  Updating an address when a member
    # changes it or the member is removed is a job for an event system, trying
    # to do that here will be very inefficient.  Currently this sort of
    # cleanup is done automatically any time a new user is subscribed or
    # unsubscribed.

    def __init__(self, context):
        super(MemberSubscriptionList, self).__init__(context)

        member_lookup = self.list_data.get('member_lookup', {})
        if isinstance(member_lookup, dict):
            self.list_data['member_lookup'] = OOBTree(member_lookup)
        self.member_lookup = self.list_data['member_lookup']

    def _getSubscribedMembers(self):
        """Return a tuple of member ids"""
        # Purge ids that no longer have an associated email in subscribers
        # as another adapter may have removed them.
        member_ids = []
        for email, info in list(self.member_lookup.items()):
            if not self.isSubscribed(email):
                del self.member_lookup[email]
            else:
                member_ids.append(info['id'])
        member_ids.sort()
        return tuple(member_ids)

    def _setSubscribedMembers(self, values):
        # BBB: ids without associated emails will be removed, perhaps they
        # should trigger a UserError instead?
        # Emails in the existing list which don't correspond to a specific
        # user name will be retained.
        new_keys = {}
        old_mapping = dict(self.member_lookup)
        for value in values:
            member = self._getMember(value)
            email = self._updateMember(member)
            if email:
                new_keys[email] = True
        # Remove emails from subscriber list which had their member ids
        # removed or email address changed
        for email in old_mapping.keys():
            if not new_keys.has_key(email):
                self.unsubscribeAddress(email)

    subscribed_members = property(fget=_getSubscribedMembers,
                                  fset=_setSubscribedMembers)

    def _updateMember(self, member):
        if member is None:
            return None

        email = self._getEmailFrom(member)

        member_id = member.getId()

        if email_regex.match(email):
            member_data = {'name': member.getProperty('fullname', member_id),
                           'id': member_id}
            # only update when things have changed
            if self.member_lookup.get(email, None) != member_data:
                # Add the members email to the subscription list and add it
                # to the member:subscription mapping.
                self.member_lookup[email] = member_data
            self.subscribeAddress(email)

        return email

    def _getEmailFrom(self, member):
        try:
            email = member.getProperty('email', None)
        except Unauthorized, e:
            try:
                # Maybe we have CMFMember, if not reraise
                email = member.getEmail()
            except AttributeError:
                raise e
        return email

    def _removeMember(self, member):

        if member is None:
            return None

        email = self._getEmailFrom(member)

        if email is not None:
            try:
                del self.member_lookup[email]
            except KeyError:
                # XXX: We should really check if it might be filed under an
                # old address
                pass
            self.unsubscribeAddress(email)

    _getSubscribers = BaseSubscriptionList._getSubscribers

    def _setSubscribers(self, value):
        value = tuple(value)
        cur_subscribers = BaseSubscriptionList._getSubscribers(self)
        if cur_subscribers != value:
            # Search through subscribed member addresses and remove those
            # that are not in the new subscriber list.
            member_addresses = self.member_lookup
            for address in member_addresses.keys():
                if address not in value:
                    del member_addresses[address]
            BaseSubscriptionList._setSubscribers(self, value)
    subscribers = property(fget=_getSubscribers, fset=_setSubscribers)

    def subscribeMember(self, user_id=None):
        new_member = self._getMember(user_id)
        self._updateMember(new_member)
        # XXX: update the whole list so we never get too out of date
        self.subscribed_members = self.subscribed_members

    def unsubscribeMember(self, user_id=None):
        member = self._getMember(user_id)
        self._removeMember(member)
        # XXX: update the whole list so we never get too out of date
        self.subscribed_members = self.subscribed_members

    def isMemberSubscribed(self):
        member = self._getMember()
        if member is not None:
            return member.getId() in self.subscribed_members
        return False

    def getSubscribedMemberFromAddress(self, address):
        return self.member_lookup.get(address, None)

    def getEmailFromId(self, user_id):
        member = self._getMember(user_id)
        if member is None: return None
        email = self._getEmailFrom(member)
        return email

    def startRecvmailMember(self, user_id=None):
        try:
            self.norecvmail.remove(self.getEmailFromId(user_id))
        except KeyError:
            pass

    def stopRecvmailMember(self, user_id=None):
        email = self.getEmailFromId(user_id)
        if email is None: return
        self.norecvmail.insert(email)

    def isMemberRecvMail(self, user_id=None):
        return self.isRecvMail(self.getEmailFromId(user_id))

    # Helper borrowed from EmailInPlone
    def _getMember(self, user_id=None):
        """Return the authenticated member, or None if anonymous.
        """
        mtool = getToolByName(self.context, 'portal_membership')
        if user_id is None:
            if mtool.isAnonymousUser():
                return None
            return mtool.getAuthenticatedMember()
        else:
            return mtool.getMemberById(user_id)


class CombinedSubscriptionList(MemberSubscriptionList):
    """An adapter implementing ICombinedSubscriptionList, requires an IAnnotable
       object to adapt.

    We again need to create a bogus MembershipTool, with some fake user info:
        >>> from Products.listen.content import tests
        >>> obj = tests.DummyAnnotableList()

    Make a dummy user and set it as the current "authenticated" user in our
    membership tool:
        >>> def_member = tests.DummyMember('dummy1', u'Dummy 1', u'dummy1@example.com')
        >>> mtool = tests.DummyMembershipTool(def_member)
        >>> mtool.dummy1 = def_member
        >>> mtool.dummy2 = tests.DummyMember('dummy2', u'Dummy 2', u'dummy2@example.com')

    We make our membership tool acquirable from the list object:
        >>> obj.portal_membership = mtool

    Make our subscription list:
        >>> clist = CombinedSubscriptionList(obj)

    Some basic tests
    ----------------
        >>> clist.subscribers
        ()
        >>> clist.all_subscribers
        ()

    Now when we set a list of emails and member names, it should just work:
        >>> clist.all_subscribers = (u'test@example.com','dummy1', 'dummy2')
        >>> clist.all_subscribers
        ('dummy1', 'dummy2', u'test@example.com')
        >>> clist.subscribers
        (u'dummy1@example.com', u'dummy2@example.com', u'test@example.com')
        >>> clist.subscribed_members
        ('dummy1', 'dummy2')

    If we set the subscribers using the methods of the original adapters our
    all_subscribers should be updated:
        >>> clist.subscribed_members = ('dummy2',)
        >>> clist.all_subscribers
        ('dummy2', u'test@example.com')
        >>> clist.subscribers = (u'dummy2@example.com', u'test2@example.com')
        >>> clist.all_subscribers
        ('dummy2', u'test2@example.com')
        >>> clist.subscribers = (u'dummy1@example.com', u'test5@example.com')
        >>> clist.all_subscribers
        (u'dummy1@example.com', u'test5@example.com')
        >>> clist.subscribed_members
        ()

    """
    implements(ICombinedSubscriptionList)

    def _getAllSubscribers(self):
        subscribers = OOSet(self._getSubscribers())
        members = OOSet(self._getSubscribedMembers())
        member_addresses = OOSet(self.member_lookup.keys())
        subscribers = difference(subscribers, member_addresses)
        subscribers = union(subscribers, members)
        return tuple(subscribers)


    def _setAllSubscribers(self, values):
        new_keys = {}
        old_subscribers = self.subscribers
        for value in values:
            try:
                member = self._getMember(value)
            except (ValueError, AttributeError, KeyError):
                member = None
            if member is None and email_regex.match(value):
                new_keys[value] = True
                self.subscribeAddress(value)
            else:
                email = self._updateMember(member)
                if email:
                    new_keys[email] = True
        # Remove emails from the lists which were removed
        for email in old_subscribers:
            if not new_keys.has_key(email):
                self.unsubscribeAddress(email)

    all_subscribers = property(fget=_getAllSubscribers,
                               fset=_setAllSubscribers)
