# Plone specific subscription adapters
import re
from AccessControl import Unauthorized
from DateTime import DateTime
from BTrees.OOBTree import OOBTree
from zope.interface import implements
from zope.i18n import translate

from plone.mail import decode_header

from Products.listen.interfaces import IMessageHandler
from Products.listen.interfaces import IInvitationSubscriptionList
from Products.listen.content.subscriptions import CombinedSubscriptionList
from Products.MailBoxer import MailBoxerTools
from Products.CMFCore.utils import getToolByName

from zope.i18nmessageid import Message
from Products.listen.i18n import _

from Products.listen.content.subscriptions import email_regex

subj_regex = re.compile('(?:mail-command:|.*: mail-command:)(\S+) (\S+)(.*)$',
                             re.IGNORECASE)

class InvitationSubscriptionList(CombinedSubscriptionList):
    """A subscription adapter which sends out invitations to those members
    and addresses that the admin wishes to subscribe.  To verify members,
    it stores a sequence of pending member subscriptions for the incoming
    request handler to verify against.

    We need to create a mailinglist and adapt it to handle
    subscriptions, we also need to ensure we have some appropriate
    fake members around::

      >>> from Products.listen.extras.tests import TestMailingList
      >>> from Products.listen.content import tests
      >>> ml = TestMailingList()
      >>> def_member = tests.DummyMember('dummy1', u'Dummy 1', u'dummy1@example.com')
      >>> mtool = tests.DummyMembershipTool(def_member)
      >>> mtool.dummy1 = def_member
      >>> ml.portal_membership = mtool
      >>> slist = InvitationSubscriptionList(ml)

    We expect an attribute that lists the pending subscriptions::

      >>> slist.pending_members
      ()

    If we add a valid member id to the list, we expect an email will
    be sent to that user (our dummy list just sets a flag)::

      >>> ml.message_count
      0
      >>> slist.pending_members = ('dummy1',)
      >>> slist.pending_members
      ('dummy1',)
      >>> ml.message_count
      1
      >>> print '\\n'.join(ml.message)
      dummy1@example.com
      mail-command:subscribe-member dummy1 [^dummy1@example.com^]
      <BLANKLINE>
      Hello Dummy 1,
      <BLANKLINE>
      You have been invited to join the MY LIST mailing list.
      If you would like to accept this invitation, simply reply to this message
      leaving the subject line unchanged.  If you do not wish to subscribe to this
      list, please disregard this message.  Any questions regarding this invitation
      should be sent to List Owner <owner@lists.example.com>.
      <BLANKLINE>
      Yours,
      List Owner <owner@lists.example.com>
      <BLANKLINE>

    When a member is subscribed to the list they are automatically removed
    from the pending queue::

      >>> slist.subscribed_members = ('dummy1',)
      >>> slist.pending_members
      ()
      >>> slist.subscribed_members
      ('dummy1',)

    Attempting to put an already subscribed member on the pending list should
    have no effect::

      >>> slist.pending_members = ('dummy1',)
      >>> slist.pending_members
      ()

    If the member is not receiving mail then they can be added to the pending list
      >>> slist.isMemberRecvMail('dummy1')
      True
      >>> slist.stopRecvmailMember('dummy1')
      >>> slist.isMemberRecvMail('dummy1')
      False
      >>> slist.pending_members = ('dummy1',)
      >>> slist.pending_members
      ('dummy1',)
      >>> slist.startRecvmailMember('dummy1')
      >>> slist.pending_members = ()


    The same is true for members without email addresses::

      >>> mtool.foobar = tests.DummyMember('foobar', '', '')
      >>> slist.pending_members = ('foobar', 'dummy1')
      >>> slist.pending_members
      ()

    Try for an email address
      >>> slist.pending_members = ('dummy2@example.com',)
      >>> slist.pending_members
      ('dummy2@example.com',)
      >>> ml.message_count
      3
      >>> print '\\n'.join(ml.message)
      dummy2@example.com
      mail-command:subscribe-member dummy2@example.com [^dummy2@example.com^]
      <BLANKLINE>
      Hello dummy2@example.com,
      <BLANKLINE>
      You have been invited to join the MY LIST mailing list.
      If you would like to accept this invitation, simply reply to this message
      leaving the subject line unchanged.  If you do not wish to subscribe to this
      list, please disregard this message.  Any questions regarding this invitation
      should be sent to List Owner <owner@lists.example.com>.
      <BLANKLINE>
      Yours,
      List Owner <owner@lists.example.com>
      <BLANKLINE>

    When an email is subscribed to the list they are automatically removed
    from the pending queue::

      >>> slist.subscribeAddress('dummy2@example.com')
      >>> slist.pending_members
      ()
      >>> slist.subscribers
      (u'dummy1@example.com', 'dummy2@example.com')

    Attempting to put an already subscribed member on the pending list should
    have no effect::

      >>> slist.pending_members = ('dummy2@example.com',)
      >>> slist.pending_members
      ()


    """



    implements(IInvitationSubscriptionList)

    _mail_body = _(u'subscription_invitation_body',u'''
Hello ${fullname},

You have been invited to join the ${listname} mailing list.
If you would like to accept this invitation, simply reply to this message
leaving the subject line unchanged.  If you do not wish to subscribe to this
list, please disregard this message.  Any questions regarding this invitation
should be sent to ${listowner}.

Yours,
${listowner}
''')

    def __init__(self, context):
        super(InvitationSubscriptionList, self).__init__(context)
        self.pending_data = self.list_data.setdefault('pending_members',
                                                      OOBTree())

    def _addPendingAddress(self, address):
        """Adds or updates an email in pending_members"""
        if address in self.subscribers:
            # XXX: If the address is already subscribed, we ignore it
            return

        # store a time stamp in case we want to invalidate invitations
        # after some interval
        self.pending_data[address] = DateTime().ISO()
        pin = self.context.pin(address)
        subject = "mail-command:subscribe-member %s [%s]"%(address, pin)
        body = self._mail_body
        mapping = { 'fullname': address,
                    'listname': self.context.title,
                    'listowner': self.context.list_owner, }
        body = Message(body, mapping=mapping)
        subject = subject.encode('ascii')
        member_mail = address.encode('ascii')
        self.context.sendCommandRequestMail(member_mail, subject,
                                            translate(body))
    


    def _addPendingMember(self, member_id):
        """Adds or updates a member in pending_members"""
        if member_id in self.subscribed_members and self.isMemberRecvMail(member_id):
            # XXX: If the user is already subscribed, we ignore it
            return
        member_info = self._getMember(member_id)
        if member_info is None:
            # If the entry is not a valid member, do nothing
            return
        try:
            member_mail = member_info.getProperty('email', None)
        except Unauthorized, e:
            try:
                # Maybe we have CMFMember, if not reraise
                member_mail = member_info.getEmail()
            except AttributeError:
                raise e
        if not member_mail:
            # if the member has no email address, do nothing
            # XXX: should perhaps raise an error for the widget here
            return
        # store a time stamp in case we want to invalidate invitations
        # after some interval
        self.pending_data[member_id] = DateTime().ISO()
        pin = self.context.pin(member_mail)
        subject = "mail-command:subscribe-member %s [%s]"%(member_id, pin)
        body = self._mail_body
        mapping = { 'fullname': member_info.getProperty('fullname', member_id),
                    'listname': self.context.title,
                    'listowner': self.context.list_owner, }
        body = Message(body, mapping=mapping)
        subject = subject.encode('ascii')
        member_mail = member_mail.encode('ascii')
        self.context.sendCommandRequestMail(member_mail, subject,
                                            translate(body))

    def _removePendingMember(self, member_id):
        """Removes a member from pending_members"""
        if self.pending_data.has_key(member_id):
            del self.pending_data[member_id]

    def _setPending(self, member_ids):
        """The mutator for the pending_members 'tuple'"""
        # Add any new members:
        for member_id in member_ids:
            if member_id not in self.pending_members:
                if email_regex.match(member_id):
                    self._addPendingAddress(member_id)
                else:
                    self._addPendingMember(member_id)
        # Remove any missing members:
        for member_id in self.pending_members:
            if member_id not in member_ids:
                self._removePendingMember(member_id)

    def _getPending(self):
        return tuple(self.pending_data.keys())

    pending_members = property(_getPending, _setPending)

    def _updateMember(self, member):
        try:
            self._updateAddress(member.getId())
        except AttributeError:
            self._updateAddress(member)
            
        return super(InvitationSubscriptionList,
                        self)._updateMember(member)

    def _updateAddress(self, address):
        self._removePendingMember(address)

    def subscribeAddress(self, email):
        if email_regex.match(email):
            self.subscriber_data.insert(email)
            self._updateAddress(email)
            try:
                self.norecvmail.remove(email)
            except KeyError:
                pass

    def _setSubscribers(self, values):
        if tuple(self.subscriber_data) != values:
            self.subscriber_data.clear()
            self.subscriber_data.update(values)
            for v in values:
                self._updateMember(v)
        for e in self.norecvmail.keys():
            if e not in self.subscriber_data:
                self.norecvmail.remove(e)

class HandleMemberSubscriptionMessages(object):
    """A simple handler for member subscription messages, it should be
    registered under the name 'subscribe-member' and will handle all messages
    with subjects containing the string 'mail-command:subscribe-member'

    We must setup our list as above::

      >>> from Products.listen.extras.tests import TestMailingList
      >>> from Products.listen.content import tests
      >>> from Products.listen.interfaces import IMailingList
      >>> from Products.listen.interfaces import IInvitationSubscriptionList
      >>> ml = TestMailingList()
      >>> def_member = tests.DummyMember('dummy1', u'Dummy 1', u'dummy1@example.com')
      >>> mtool = tests.DummyMembershipTool(def_member)
      >>> mtool.dummy1 = def_member
      >>> ml.portal_membership = mtool

    This adapter needs to interact with an adapter providing
    IInvitationSubscriptionList, s we must register one::

      >>> from zope.component import provideAdapter
      >>> provideAdapter(InvitationSubscriptionList, [IMailingList], IInvitationSubscriptionList)

    We also need a fake request in which to embed our mail message, for
    this case a simple mapping should suffice::

      >>> request = {}
      >>> request['Mail'] = ('''To: list@list.com
      ... From: Dummy 1 <dummy1@example.com>
      ... Subject: Re: mail-command:subscribe-member dummy1
      ... 
      ... A message.
      ... ''')

    Now we adapt our list and request to handle this incoming mail,
    this request mail includes no pin and simply requests a
    subscription for the above member.  Let's see if it works::

      >>> handler = HandleMemberSubscriptionMessages(ml, request)
      >>> ml.message_count
      0
      >>> handler.subscriber_info.pending_members
      ()
      >>> handler.subscriber_info.subscribed_members
      ()
      >>> handler.processMail()

    Nobody should have been subscribed but an invitation should have
    been sent (becuse the from address matched the member address)::

      >>> ml.message_count
      1
      >>> handler.subscriber_info.pending_members
      (u'dummy1',)
      >>> handler.subscriber_info.subscribed_members
      ()

    Now we are on the pending list, we can check what happens when we
    send a message with the expected pin in the subject::

      >>> subject = 'Subject: Re: ' + ml.message[1]
      >>> subject_re = re.compile('^Subject:.*$', re.MULTILINE)
      >>> request['Mail'] = subject_re.sub(subject, request['Mail'])
      >>> handler = HandleMemberSubscriptionMessages(ml, request)
      >>> handler.processMail()
      >>> ml.message_count
      1
      >>> ml.confirm_count
      1
      >>> ml.confirm
      ['dummy1@example.com', 'Dummy 1']
      >>> handler.subscriber_info.subscribed_members
      ('dummy1',)
      >>> handler.subscriber_info.pending_members
      ()

    If we are already subscribed to the list, then sending a
    subscription request should result in a bounce message indicating
    the issue::

      >>> handler.processMail()
      >>> ml.message_count
      1
      >>> ml.confirm_count
      1
      >>> ml.bounce_count
      1
      >>> ml.bounce
      ['dummy1@example.com', u'Re: mail-command:subscribe-member dummy1 [^dummy1@example.com^]']

    If already subscribed but user is not receiving mail, then message
    should not be bounced but user should be removed from norecvmail
      >>> handler.subscriber_info.stopRecvmailMember('dummy1')
      >>> handler.subscriber_info.isMemberRecvMail('dummy1')
      False
      >>> handler.processMail()
      >>> ml.message_count
      2


    """

    implements(IMessageHandler)

    def __init__(self, context, request):
        self.context = context
        self.request = request
        # get subscription info
        self.subscriber_info = IInvitationSubscriptionList(self.context)

        message = self.request.get('Mail')
        (header, body) = MailBoxerTools.splitMail(message)
        # get email-address
        sender = decode_header(header.get('from',''))
        (name, email) = MailBoxerTools.parseaddr(sender)
        # Because of MailBoxer's naive ASCII only response messages we
        # brute force some ascii here
        self.name = name.encode('ascii', 'replace')
        self.email = email.encode('ascii', 'replace')
        self.pin = self.context.pin(self.email)
        # get subject
        self.subject = decode_header(header.get('subject', ''))

    def processMail(self):
        """Take an incoming message and try to verify it against a registry
        of pending member subscriptions using a PIN. If it validates, subscribe
        the new member"""
        # Split the subject into the expected command paramters
        param_match = subj_regex.match(self.subject)
        if not param_match:
            # XXX: If the subject is not properly formatted, ignore the message
            return
        (command, user_id, pin_stuff) = param_match.groups()
        if self.email in self.subscriber_info.subscribers and self.email not in self.subscriber_info.norecvmail:
            # XXX: This method is not in the interface
            self.context.sendBounceTo(self.email, self.name, self.subject)
            return

        expected_pin = self.context.pin(self._getEmail(user_id))

        pending_emails = map(self._getEmail, self.subscriber_info.pending_members)

        if self.email in pending_emails:
            # Ensure the sender and user have matching mails, and that the
            # corresponding pin is in the message subject
            if expected_pin == self.pin and pin_stuff.find(self.pin) != -1:
                # subscribe the member if the pin matches
                if user_id not in pending_emails:
                    self.subscriber_info.subscribeMember(user_id)
                    if not self.subscriber_info.isRecvMail(self.email):
                        self.subscriber_info.norecvmail.remove(self.email)
                else:
                    self.subscriber_info.pending_members = \
                        [x for x in self.subscriber_info.pending_members if x != self.email]
                    self.subscriber_info.subscribeAddress(self.email)
                
                # XXX: This isn't in the interface, but should be acceptable
                self.context.sendSubscribeConfirmationFor(self.email, self.name)
                return
        if expected_pin == self.pin:
            # This is a request without a pin or a request for a non-pending
            # member, add the member to the pending list and resend provided
            # that the sender address and the member address match.
            self._resendRequest(user_id)
        # XXX: what do we do if the email for the request doesn't match the
        # email for the user?

    def _getEmail(self, user_id):
        if email_regex.match(user_id): return user_id
        
        member = self._getMember(str(user_id))

        if member is None: raise ValueError(user_id)

        try:
            return member.getProperty('email', None)
        except Unauthorized, e:
            try:
                # Maybe we have CMFMember, if not reraise
                return member.getEmail()
            except AttributeError:
                raise e


    def _getMember(self, user_id):
        mtool = getToolByName(self.context, 'portal_membership')
        return mtool.getMemberById(user_id)

    def _resendRequest(self, user_id):
        """Remove the member from the pending list and readd, so
        a proper invitation will be sent (in case the user was
        simply requesting an invitation)"""
        self.subscriber_info.pending_members = tuple([m for m in
                                            self.subscriber_info.pending_members
                                                      if m != user_id])
        self.subscriber_info.pending_members = (
                                      self.subscriber_info.pending_members +
                                      (user_id,))
