from zope.interface import Interface
from zope.app.annotation.interfaces import IAnnotatable
from zope.schema import TextLine
from zope.schema import Text
from zope.schema import ASCII
from zope.schema import Choice
from zope.schema import Bool
from zope.schema import Tuple

from Products.listen.i18n import _

class IMailingList(Interface):
    """
       A de facto interface for some MailBoxer-y things, includes the basic
       public features, minus those that are useful primarily for
       smtp2zope.py
    """

    title = TextLine(
        title = _(u"Title"),
        description = _(u""),
        required = True,)

    description = Text(
        title = _(u"Description"),
        description = _(u"A description of the mailing list."),
        default=u'',
        required = False)

    mailto = ASCII(
        title = _(u"List Address"),
        description = _(u"Main address for the mailing list."),
        required = True,)

    list_owner = TextLine(
        title = _(u"Owner Address"),
        description = _(u"The email address of the list owner.  The owner is "
                         "always a moderator if moderation is enabled.  The "
                         "owner is also the recipient of any bounced mails."),
        required = True,)

    moderators = Tuple(
        title = _(u"Moderators"),
        description = _(u"The email addresses to send moderation requests to, "
                         "when moderation is enabled."),
        default = (),
        required = False,
        value_type = TextLine(title=_(u"Moderator"),),)

    moderated = Bool(
        title = _(u"Moderated"),
        description = _(u"When moderation is enabled, all messages sent to "
                          "the list will not be distributed until approved "
                          "by a moderator."),
        default = False,
        required = False,)

    archived = Choice(
        title = _(u"Archival method"),
        description = _(u"When archiving is enabled, all messages sent to "
                         "the list will be saved on the server.  You may "
                         "choose whether to archive just the message text, "
                         "or include attachments."),
        vocabulary = "Archive Options",
        required = True,
        )

    closed = Bool(
        title = _(u"Closed List"),
        description = _(u"A list that is closed will only accept mail from "
                         "subscribed addresses."),
        default = True,
        required = False,)

# These methods really belong in an adapter(s), but because we are using
# MailBoxer they are part of the content object.

    def addMail(message_string):
        """
           Parses a mail message into a string and stores it in the archive
           as an IMailMessage.  Returns the resulting mail message object.
        """

    def checkMail(REQUEST):
        """
           Extracts a message from an HTTP request checks its validity.
           Returns None if the mail passes validation, otherwise it returns
           a string describing the error.  Expects a request variable 'Mail'
           containing the mail message with headers.
        """

    def requestMail(REQUEST):
        """
           Extracts an (un)subscribe request message from an HTTP request and
           processes it.  Expects a request variable 'Mail' containing the
           mail message with headers.
        """

    def processMail(REQUEST):
        """
           Extracts a message from an HTTP request and processes it for the
           mailing list.  Checks the validity of the sender and whether
           moderation is needed.  Expects a request variable 'Mail' containing
           the mail message with headers
        """

    def moderateMail(REQUEST):
        """
           Processes an HTTP request for information on what to do with a
           specific mail message in the moderation queue.  Expects the request
           to contain an 'action' ('approve' or 'discard'), a 'pin' used to
           authenticate the moderator, and an 'mid' which is the id of the
           mail object in the moderation queue.  Expects a request variable
           'Mail' containing the mail message with headers
        """

    def bounceMail(REQUEST):
        """
           Extracts a bounce message from an HTTP request and processes it to
           note any subscriber addresses that resulted in bounces.
        """

    def resetBounces(addresses):
        """
           Remove specified email addresses from the list of recipients with
           bounced messages.
        """

    def sendCommandRequestMail(address, subject, body):
        """
           Sends an administrative mail to the user requesting a response
           so that list commands can be executed.
        """

# XXX: this method needs to be public for now, so it does not belong in the
# interface
#    def pin(email):
#        """
#           Generates a pin from an email address, using an internal hash
#        """
#
# XXX: Can't add this to the interface because that causes it to become
# non-public, and individual method permission declarations don't seem to be
# working in Five
#     def uniqueMessageId():
#         """
#             Generate a message id for messages which are missing one.
#         """


class ISubscriptionList(Interface):
    """ An interface for (un)subscribing mail addresses to a list. """

    subscribers = Tuple(
        title = _(u"Subscribers"),
        description = _(u"The email addresses subscribed to the list."),
        required = False,
        value_type = TextLine(title=_(u"Subscriber"),),)

    def deleteBounces(addresses):
        """
           Remove specified email addresses from the subscriber list and
           remove them from the list of bounced recipients.
        """

    def subscribeAddress(email):
        """
           Add an email to the list of subscribers.
        """

    def unsubscribeAddress(email):
        """
           Remove an email from the list of subscribers.
        """

    def isSubscribed(email):
        """
           Returns boolean indicating whether the address is subscribed.
        """
    
    def isRecvMail(email):
        """
           Returns boolean indicating whether the address is receiving mail.
        """

    def getAllowedSenders():
        """
           Returns all allowed senders (includes those subscribed but not
           receiving mail)
        """
    
    def stopRecvmail(email):
        """
           Sets the email address to not receive mail
        """


class IMemberSubscriptionList(ISubscriptionList):
    """ 
       An interface for (un)subscribing and changing the receive mail status
       of members on a list. 
    """

    subscribed_members = Tuple(
        title = _(u"Subscribed Members"),
        description = _(u"The ids of members subscribed to the list."),
        required = False,
        value_type = TextLine(title=_(u"Member"),),)

    def subscribeMember(user_id=None):
        """
           Add the specified user_id to the mailing list.  If None is
           specified the current member is assumed.
        """

    def unsubscribeMember(user_id=None):
        """
           Remove member from the list of subscribers.  If None is specified
           the current member is assumed.
        """

    def isMemberSubscribed(user_id=None):
        """
           Returns boolean indicating whether the member is subscribed.  If
           None is specified the subscription status of the current user is
           checked.
        """

    def getSubscribedMemberFromAddress(address):
        """
            Utility method for linking messages to members.  Returns a mapping
            containing the userid ('id') and the full name ('name') of the
            subscribed member with the given address.
        """

    def startRecvmailMember(user_id=None):
        """
           Sets the member to receive mail
        """

    def stopRecvmailMember(user_id=None):
        """
           Sets the member to not receive mail
        """

    def isMemberRecvMail(user_id=None):
        """
           Returns boolean indicating whether the member is receiving mail.  If
           None is specified the receiving mail status of the current user is
           checked.
        """

class ICombinedSubscriptionList(IMemberSubscriptionList):
    """A convenience interface for emails and member ids at the same time"""

    all_subscribers = Tuple(
        title = _(u"All Subscribers"),
        description = _(u"The ids of all members/emails who are allowed to send mails"),
        required = False,
        value_type = TextLine(title=_(u"User id or email address"),),)

class IInvitationSubscriptionList(ICombinedSubscriptionList):
    """A subscription adapter which allows a list manager to select members
    or addresses to invite to the list.  When invited, those
    members go onto a list of pending addresses while awaiting
    verification from the intended subscriber."""

    pending_members = Tuple(
        title = _(u"Pending Members"),
        description = _(u"The ids of pending member subscribers."),
        required = False,
        value_type = TextLine(title=_(u"User id"),),)


class IHaveSubscribers(IAnnotatable):
    """
        A marker interface indicating that an object wants to provide
        subscribers
    """

class IListArchive(Interface):
    """
        A marker interface to indicate that a folder contains a list archive.
    """


class IListArchiveSubFolder(IListArchive):
    """
        A marker interface to indicate that a folder is part of a list archive.
    """


class IModerationQueue(Interface):
    """
        A marker interface to indicate that a folder contains a moderation
        queue.
    """
